from jinja2 import Environment, FileSystemLoader

from . import join, root_path
from .config import jinja_options
from .context import context, filters
from .util import _


options = _(jinja_options)
options.loader = FileSystemLoader(join(root_path, 'templates'))

templates = Environment(**options)

templates.hamlish_enable_div_shortcut = True
templates.hamlish_mode = 'indented'

templates.filters.update(filters)
templates.globals.update(context)



# -- Project information -----------------------------------------------------

project = 'junior'
copyright = '2020, Chris Whalen'
author = 'Chris Whalen'

# The full version, including alpha/beta/rc tags
release = '0.1.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['static']



from flask_sqlalchemy import SQLAlchemy

from sqlalchemy import event
from sqlalchemy.orm import synonym, validates

from . import dt
from .errors import InvalidRequestError


def model(this):

    @event.listens_for(this, 'init')
    @event.listens_for(this, 'load')
    def load(self, *args, **params):

        self.session = db.create_scoped_session(options={'bind': None})

        try:
            db.session.expunge(self)

        except InvalidRequestError:
            pass

        self.session.add(self)

    def delete(self):

        self.session.delete(self)
        self.session.commit()
        return self

    def fill(self, **params):

        for param in params:
            setattr(self, param, params[param])

        return self

    def save(self):

        if not self.id:
            self.session.add(self)

        self.session.commit()
        return self

    this.delete = delete
    this.fill = fill
    this.save = save

    return this


def on(_event, attribute, function):

    def wrapper(this):

        @event.listens_for(getattr(this, attribute), event)
        def listen(target, value, source):

            try:
                return function(value, target, source)

            except TypeError:

                try:
                    return function(value, target)

                except TypeError:

                    try:
                        return function(value)

                    except TypeError:
                        return function()

        return this

    return wrapper


def filter(_event, attribute, function):

    def wrapper(this):

        @event.listens_for(getattr(this, attribute),
                           _event, retval=True)
        def listen(target, value, oldvalue, source):

            try:
                return function(value, oldvalue, target, source)

            except TypeError:

                try:
                    return function(value, oldvalue, target)

                except TypeError:

                    try:
                        return function(value, oldvalue)

                    except TypeError:

                        try:
                            return function(value)

                        except TypeError:
                            return function()

        return this

    return wrapper


db = SQLAlchemy()


class WithTimestamps:

    created_at = db.Column(db.DateTime, default=dt.now)
    updated_at = db.Column(db.DateTime, default=dt.now, onupdate=dt.now)



from flask_security import current_user

from .errors import MultipleResultsFound, NoResultFound
from .models import User


def access():

    permit(User,
           get=deny_all
           )

    permit(User,
           get=allow_self_for_current_user,
           put=allow_self_for_current_user,
           patch=allow_self_for_current_user
           )


def deny_all(model):

    return None


def allow_all(model):

    return all


def allow_self_for_current_user(model):

    try:
        return User.query.filter_by(id=current_user.id)

    except (AttributeError, MultipleResultsFound, NoResultFound):
        return None


def permit(model, **params):

    for p in params:

        q = params[p](model)
        action = params[p].__name__.split('_')[0]

        if action == 'allow':

            if (model.resource.Meta.allowed[p] is all) or (q is all):
                model.resource.Meta.allowed[p] = all

            elif model.resource.Meta.allowed[p] is None:
                model.resource.Meta.allowed[p] = q

            elif q is not None:
                model.resource.Meta.allowed[p] = \
                    model.resource.Meta.allowed[p].union(q)

        if action == 'deny':

            if (model.resource.Meta.allowed[p] is None) or (q is None):
                model.resource.Meta.allowed[p] = None

            elif model.resource.Meta.allowed[p] is all:
                model.resource.Meta.allowed[p] = q

            elif q is not all:
                model.resource.Meta.allowed[p] = \
                    model.resource.Meta.allowed[p].intersect(q)



from re import sub

from flask_marshmallow import Marshmallow

from flask_restful import Api as RestfulApi, Resource

from . import Blueprint, _
from .config import config
from .controls import permit
from .errors import (
    DataError, MultipleResultsFound, NoResultFound, NotFound, error, handle)


class Api(RestfulApi):

    def _should_use_fr_error_handler(self):
        pass

    def error_router(self, handler, e):
        return handler(e)


api = Blueprint('api', __name__)
rest = Api(api)
schemas = Marshmallow()
resources = []

properties = _(
    endpoints={},
    root='/api/v%i/' % (config.api.version,),
    version=config.api.version,
    updated_at=config.api.updated_at
)

api.url_prefix = properties.root[:-1]


def resource(this):

    class ModelSchema(schemas.SQLAlchemyAutoSchema):

        class Meta:

            model = this
            exclude = []
            only = []

            try:
                exclude = this.Meta.exclude

            except AttributeError:
                pass

            try:
                only = this.Meta.only

            except AttributeError:
                pass

    class ModelResource(Resource):

        class Meta:

            model = this
            allow = _(query=all, add=None, update=None, delete=None)

            try:
                endpoint = this.Meta.endpoint

            except AttributeError:
                pass

        def query(self):

            if self.Meta.allow.query is all:
                return self.Meta.model.query

            if self.Meta.allow.query is None:
                return self.Meta.model.query.limit(0).from_self()

            return self.Meta.allow.query.from_self()

        def get(self, **params):

            try:
                query = self.query()

                if 'id' in params:
                    return self.Meta.model.one.jsonify(
                        query.filter_by(**params).one())

                return self.Meta.model.many.jsonify(query.all())

            except NoResultFound:
                return error(404)

            except (DataError, MultipleResultsFound):
                return error(400)

        def post(self, **params):

            if self.Meta.allow.add is all:

                try:
                    record = self.Meta.model()
                    record.fill(**params)
                    record.save()
                    return self.Meta.model.one.jsonify(record)

                except DataError:
                    return error(400)

            return {}

        def put(self, **params):

            if self.Meta.allow.update is None:
                return {}

            try:
                if 'id' in params:

                    record = self.Meta.model.get(params['id'])

                    if (self.Meta.allow.update is all or
                            record in self.Meta.allow.update):

                        record.fill(**params)
                        record.save()
                        return self.Meta.model.one.jsonify(record)

                else:
                    return error(400)

            except DataError:
                return error(400)

            return {}

        def patch(self, **params):

            if self.Meta.allow.update is None:
                return {}

            try:
                if 'id' in params:

                    record = self.Meta.model.get(params['id'])

                    if (self.Meta.allow.update is all or
                            record in self.Meta.allow.update):

                        record.fill(**params)
                        record.save()
                        return self.Meta.model.one.jsonify(record)

                else:
                    return error(400)

            except DataError:
                return error(400)

            return {}

        def delete(self, **params):

            if self.Meta.allow.delete is None:
                return {}

            try:
                if 'id' in params:

                    record = self.Meta.model.get(params['id'])

                    if (self.Meta.allow.delete is all or
                            record in self.Meta.allow.delete):

                        record.delete()
                        return self.Meta.model.one.jsonify(record)

                else:
                    return error(400)

            except DataError:
                return error(400)

            return {}

    ModelResource.__name__ = '%sResource' % (this.__name__,)
    this.resource = ModelResource

    ModelSchema.__name__ = '%sSchema' % (this.__name__,)
    this.one = ModelSchema(many=False)
    this.many = ModelSchema(many=True)
    this.schema = ModelSchema

    return this


def register(this, name=None, endpoint=None, *params):

    try:
        resource = this.resource

    except AttributeError:
        resource = this

    if name is None:

        name = resource.__name__

        if name.endswith('Resource') and name != 'Resource':
            name = name[:-8]

    if endpoint is None:

        try:
            endpoint = resource.Meta.endpoint

        except AttributeError:
            endpoint = '%ss' % (sub(r'([A-Z])', r'_\1', name)[1:].lower(),)

    paths = ['/%s' % (endpoint,)]
    properties.endpoints[name] = '%s%s/' % (properties.root, endpoint)

    if len(params) == 0:

        try:
            params = ([key.name for key in
                      resource.Meta.model.__table__.primary_key])

        except AttributeError:
            pass

    for param in params:
        paths.append('%s/<string:%s>' % (paths[-1], param))

    if len(paths) > 1:
        paths[0] = '%s/' % paths[0]

    else:
        properties.endpoints[name] = properties.endpoints[name][:-1]

    rest.add_resource(resource, *paths)
    resources.append(resource)

    return this


def defaults():

    @api.route('/')
    def __index():

        return properties

    @api.route('/<path:path>')
    def __default(**params):

        raise NotFound

    @handle(api.url_prefix+'/')
    @api.errorhandler(Exception)
    def __error(exception):

        message = error(exception)
        return {'error': message}, message.code


def start(app):

    from . import _api

    _api.defaults()

    with app.app_context():

        for resource in _api.resources:

            try:
                permit(resource.Meta.model, **resource.Meta.model.Meta.control)

            except AttributeError:
                pass

    app.register_blueprint(_api.api,
                           url_prefix=_api.api.url_prefix)

    return app



from . import Blueprint, redirect, render
from .api import properties
from .context import context
from .errors import error, handle
from .templates import templates


web = Blueprint('web', __name__)


def redirect_to_api(path, api_path='api'):

    if path == api_path:
        return redirect(properties.root)

    if path.split('/')[0] == api_path:
        return redirect(''.join([properties.root,
                        path.replace(api_path+'/', '')]))


def defaults():

    @web.route('/favicon<name>')
    def __favicon(name):

        return redirect(context.asset('images/favicon%s' % (name,)))

    @web.route('/')
    def __index():

        try:
            return render('index.haml')

        except Exception:
            return templates.get_template('index.haml').render()

    @web.route('/<path:path>')
    def __default(**params):

        if ('path' in params and
                params['path'] == 'api' or
                params['path'].startswith('api/')):

            return redirect_to_api(params['path'])

        try:
            return render('index.haml')

        except Exception:
            return templates.get_template('index.haml').render()

    @handle('/')
    @web.errorhandler(Exception)
    def __error(exception):

        e = error(exception)

        try:
            return render('error.haml', **e), e.code

        except Exception:
            return templates.get_template('index.haml').render(**e), e.code


def start(app):

    from . import _web

    _web.defaults()

    app.register_blueprint(_web.web)

    return app



from celery import Celery

from .config import celery_options


queue = Celery('app')
queue.conf.update(celery_options)


def start(app):

    class Task(queue.Task):

        def __call__(self, *args, **params):

            with app.app_context():
                return self.run(*args, **params)

    queue.Task = Task

    return app



from junior.cache import cache, clear, delete, forget, get, memo, set, store


class TestCache:

    # Does the cache set and get simple values?
    def test_set_get_simple(self):

        set('a nice number', 42)

        assert get('a nice number') == 42

    # Does the cache set and get unicode strings?
    def test_set_get_unicode(self):

        set('a fancy string', '◎ܫ◎ ಠﭛಠ ⊙_ʘ ♨_♨ ಠ_ಠ')

        assert get('a fancy string') == '◎ܫ◎ ಠﭛಠ ⊙_ʘ ♨_♨ ಠ_ಠ'

    # Does the cache set and get lists?
    def test_set_get_list(self):

        set('a short list', ['one', 'two', 'three'])

        assert get('a short list') == ['one', 'two', 'three']

    # Does the cache set and get dicts?
    def test_set_get_dict(self):

        set('a little dict', {'un': 1, 'deux': 2, 'trois': 3})

        assert get('a little dict') == {'un': 1, 'deux': 2, 'trois': 3}

    # Does the cache delete values?
    def test_delete(self):

        set('baby', 'boy')

        assert get('baby') == 'boy'

        delete('baby')

        assert get('baby') is None

    # Does the cache clear?
    def test_clear(self):

        set('once upon', 'a time')
        set('i visited', 'the moon')

        assert get('once upon') == 'a time'
        assert get('i visited') == 'the moon'

        clear()

        assert get('once upon') is None
        assert get('i visited') is None



from flask_security import current_user

from .errors import MultipleResultsFound, NoResultFound


def deny_all(model):

    return None


def allow_all(model):

    return all


def allow_self_for_current_user(model):

    try:
        return model.query.filter_by(id=current_user.id)

    except (AttributeError, MultipleResultsFound, NoResultFound):
        return None


def allow_belongs_to_current_user(model):

    try:
        return model.query.filter_by(user_id=current_user.id)

    except (AttributeError, MultipleResultsFound, NoResultFound):
        return None


def permit(model, **params):

    for p in params:

        controls = params[p]

        try:
            controls[0]

        except TypeError:
            controls = (controls,)

        for control in controls:

            query = control(model)
            action = control.__name__.split('_')[0]

            if action == 'allow':

                if (model.resource.Meta.allow[p] is all) or (query is all):
                    model.resource.Meta.allow[p] = all

                elif model.resource.Meta.allow[p] is None:
                    model.resource.Meta.allow[p] = query

                elif query is not None:
                    model.resource.Meta.allow[p] = \
                        model.resource.Meta.allow[p].union(query)

            if action == 'deny':

                if (model.resource.Meta.allow[p] is None) or (query is None):
                    model.resource.Meta.allow[p] = None

                elif model.resource.Meta.allow[p] is all:
                    model.resource.Meta.allow[p] = query

                elif query is not all:
                    model.resource.Meta.allow[p] = \
                        model.resource.Meta.allow[p].intersect(query)



from re import compile

from jinja2.exceptions import TemplateSyntaxError
from jinja2.ext import Extension
from jinja2.lexer import Token, count_newlines


class HamlBabel(Extension):

    inner = compile(r'\\?[()]')

    outer = compile(r'\\?(_|__|gettext|ngettext|lazy_gettext|lazy_ngettext)\(')

    def filter_stream(self, stream):

        parens = 0
        method = None

        for token in stream:

            if token.type != 'data':

                yield token
                continue

            line = token.lineno
            pos = 0

            while True:

                if not parens:
                    match = self.outer.search(token.value, pos)

                else:
                    match = self.inner.search(token.value, pos)

                if match is None:
                    break

                new_pos = match.start()

                if new_pos > pos:

                    parsed = token.value[pos:new_pos]

                    if method in ('__', 'ngettext', 'lazy_ngettext'):

                        segments = parsed.split('||')

                        # print(segments)

                        yield Token(line, 'data', segments[0])
                        yield Token(line, 'block_begin', None)
                        yield Token(line, 'name', 'pluralize')
                        yield Token(line, 'block_end', None)

                        for segment in segments[1:]:

                            yield Token(line, 'data', segment)

                    else:
                        yield Token(line, 'data', parsed)

                    line += count_newlines(parsed)

                group = match.group()

                if group[0] == '\\':
                    yield Token(line, 'data', group[1:])

                elif not parens:

                    method = group[:-1]

                    yield Token(line, 'block_begin', None)
                    yield Token(line, 'name', 'trans')
                    yield Token(line, 'block_end', None)
                    parens = 1

                else:
                    if group == '(' or parens > 1:
                        yield Token(line, 'data', group)

                    parens += group == ')' and -1 or 1

                    if not parens:
                        yield Token(line, 'block_begin', None)
                        yield Token(line, 'name', 'endtrans')
                        yield Token(line, 'block_end', None)

                pos = match.end()

            if pos < len(token.value):

                # print(token.value)
                yield Token(line, 'data', token.value[pos:])

        if parens:

            raise TemplateSyntaxError(
                'Unmatched parentheses in gettext expression',
                token.line, stream.name, stream.filename)



from os.path import split
from re import sub

from jinja2 import Environment

from webassets.filter import Filter, register_filter as register

from . import join
from .config import jinja_options, vendor
from .context import context


@register
class Hamlish(Filter):

    name = 'hamlish'

    def input(self, _in, _out, **params):

        _out.write(Environment(**jinja_options).hamlish_from_string(
            _in.read(), globals=context).render())


@register
class HandleEmpty(Filter):

    name = 'handle_empty'

    def input(self, _in, _out, **params):

        style = _in.read()

        if not len(style.strip()):
            style = '/* */'

        _out.write(style)


@register
class Require(Filter):

    name = 'require'

    def require(self, match):

        missing = False

        path = list(split(match.groups()[-1]))
        path[-1] = '%s.js' % (path[-1],)

        prefix_path = [p for p in path]
        prefix_path[-1] = '_%s' % (prefix_path[-1],)

        try:
            return open(join(self.path, *prefix_path), 'r').read()

        except FileNotFoundError:

            try:
                return open(join(self.path, *path), 'r').read()

            except FileNotFoundError:

                name = match.groups()[0]
                path = join(self.ctx.app.root_path, 'node_modules', name)

                try:
                    return open(join(path, vendor.scripts[name]), 'r').read()

                except (FileNotFoundError, KeyError, TypeError):

                    try:
                        return open(join(path, '%s.js' % (name,)), 'r').read()

                    except FileNotFoundError:

                        try:
                            return open(join(path, 'dist', '%s.js' % (name,)),
                                        'r').read()

                        except FileNotFoundError:

                            try:
                                return open(join(path, '%s.min.js' % (name,)),
                                            'r').read()

                            except FileNotFoundError:

                                try:
                                    return open(join(path, 'dist',
                                                     '%s.min.js' % (name,)),
                                                'r').read()

                                except FileNotFoundError:

                                    missing = True

        if missing:
            raise FileNotFoundError("require('%s')" % (name,))

    def input(self, _in, _out, **params):

        self.path = split(params['source_path'])[0]
        for line in _in.readlines():

            _out.write(sub(r'^\s*require\(\s*["\'](.*?)["\']\s*\)\s*;?',
                           self.require,
                           line))


@register
class ImportSSS(Filter):

    name = 'import_sss'

    def _import(self, match):

        name = list(split(match.groups()[0]))[-1]

        path = '%s.sss' % (name,)
        prefix_path = '_%s' % (path,)

        try:
            return open(join(self.path, prefix_path), 'r').read()

        except FileNotFoundError:

            try:
                return open(join(self.path, path), 'r').read()

            except FileNotFoundError:

                return "@import '%s'" % (name,)

    def input(self, _in, _out, **params):

        self.path = split(params['source_path'])[0]
        for line in _in.readlines():

            _out.write(sub(r'^\s*@import\s+["\'](.*?)["\']\s*[;\n]',
                           self._import,
                           line))


@register
class ImportCSS(Filter):

    name = 'import_css'

    def _import(self, match):

        missing = False

        name = match.groups()[0]
        path = join(self.ctx.app.root_path, 'node_modules', name)

        try:
            return open(join(path, vendor.styles[name]), 'r').read()

        except (FileNotFoundError, KeyError):

            try:
                return open(join(path, '%s.css' % (name,)), 'r').read()

            except FileNotFoundError:

                try:
                    return open(join(path, 'dist', '%s.css' % (name,)),
                                'r').read()

                except FileNotFoundError:

                    try:
                        return open(join(path, '%s.min.css' % (name,)),
                                    'r').read()

                    except FileNotFoundError:

                        try:
                            return open(join(path, 'dist',
                                             '%s.min.css' % (name,)),
                                        'r').read()

                        except FileNotFoundError:

                            missing = True

        if missing:
            raise FileNotFoundError("@import '%s'" % (name,))

    def input(self, _in, _out, **params):

        self.path = split(params['source_path'])[0]
        for line in _in.readlines():

            _out.write(sub(r'^\s*@import\s+["\'](.*?)["\']\s*[;\n]',
                           self._import,
                           line))


@register
class FixBabelHelpers(Filter):

    name = 'fix_babel_helpers'

    def fix_babel_helpers(self, match):

        return ''

    def input(self, _in, _out, **params):

        for line in _in.readlines():

            _out.write(sub(r'babelHelpers\["(.*?)"\]',
                           self.fix_babel_helpers,
                           line))


@register
class Strip(Filter):

    name = 'strip'

    def output(self, _in, _out, **params):

        for line in _in.readlines():

            if len(line.strip()):
                _out.write(line)


@register
class WrapScript(Filter):

    name = 'wrap_script'

    def output(self, _in, _out, **params):

        script = _in.read()

        if len(script.strip()) > len('"use strict";'):
            _out.write('<script async>\n%s\n</script>' % (script,))


@register
class WrapStyle(Filter):

    name = 'wrap_style'

    def output(self, _in, _out, **params):

        style = _in.read()

        if len(style.strip()):
            _out.write('<style>\n%s\n</style>' % (style,))



from os import system

import click

from flask.cli import with_appcontext

import konch

from .config import env
from .util import dump


@click.group()
def locale():
    """Extract and compile translations."""


@locale.command('extract')
def extract_locale():
    """Extract messages into a POT template."""
    system('pybabel extract -F %s/babel.cfg -k X -o messages.pot .'
           % (env.cache_dir,))


@locale.command('create')
@click.argument('name')
def create_locale(name):
    """Create a new locale."""
    system('pybabel init -i messages.pot -d translations -l %s' % (name,))


@locale.command('update')
def update_locale():
    """Update existing locales to match the current POT template."""
    system('pybabel update -i messages.pot -d translations')


@locale.command('compile')
def compile_locale():
    """Compile all locales."""
    system('pybabel compile -d translations')


@click.group()
def queue():
    """Manage the task queue."""


@queue.command('run')
@click.argument('name', required=False)
@with_appcontext
def run_queue(name):
    """Start the task worker.

    NAME is the name of your application's queue module. If not provided,
    this defaults to the "queue" property of your app module.
    """

    from flask.globals import _app_ctx_stack
    app = _app_ctx_stack.top.app

    if (name is None):
        name = '%s.queue' % (app.import_name,)

    system('celery worker -E -A %s' % (name,))


@click.command('run')
@click.argument('name', required=False)
@click.option('-b', '--bind', default='127.0.0.1:8000', show_default=True,
              help='Bind to address:port.')
@click.option('-w', '--workers', default=4, show_default=True,
              help='Number of worker processes.')
@click.option('-r', '--reload', is_flag=True, default=False,
              show_default=True, help='Restart the server when code changes.')
@click.option('-p', '--preload', is_flag=True, default=False,
              show_default=True,
              help='Preload the application before starting the server.')
@click.option('--key', help='Path the the SSL private key.')
@click.option('--cert', help='Path the the SSL certificate.')
@click.option('--ca', help='Path the the SSL certificate authority (CA).')
@with_appcontext
def run(name, bind, workers, reload, preload, key, cert, ca):
    """Start a gunicorn server.

    NAME is the name of your application module. If not provided,
    it will be autodetected.
    """

    from flask.globals import _app_ctx_stack
    app = _app_ctx_stack.top.app

    if (name is None):
        name = app.import_name

    options = ''

    if (reload):
        options += '--reload '

    if (preload):
        options += '--preload '

    if (key):
        options += '--keyfile %s ' % (key,)

    if (cert):
        options += '--certfile %s ' % (cert,)

    if (ca):
        options += '--ca-certs %s ' % (ca,)

    system('gunicorn -b %s -w %s -n %s %s %s' %
           (bind, workers, name, options, name))


@click.command()
@with_appcontext
def shell():
    """Start a shell in the app context."""

    def context_format(ctx):

        return ', '.join(sorted(ctx.keys()))

    def shell_dump(*args):

        print(dump(*args, _print=False))

    from flask.globals import _app_ctx_stack
    app = _app_ctx_stack.top.app

    shell_context = {'app': app}

    imports = ('Application', 'Blueprint', 'Flask', 'Request', 'Resource',
               'Response', 'X', '_', 'api', 'cache', 'components', 'config',
               'context', 'db', 'dt', 'emit', 'env', 'error', 'forget', 'g',
               'join', 'jsonify', 'memo', 'model', 'on', 'queue', 'redirect',
               'register', 'render', 'request', 'resource', 'response',
               'schemas', 'send', 'session', 'socket', 'split', 'store', 'to',
               'web')

    import junior

    for name in imports:
        shell_context[name] = getattr(junior, name)

    shell_context.update(app.make_shell_context())
    shell_context['dump'] = shell_dump

    konch.start(shell='auto',
                context=shell_context,
                context_format=context_format,
                banner=click.style('%s :: %s\n' %
                                   (app.import_name, app.env), bold=True),
                )


def start(app):

    app.cli.add_command(locale)
    app.cli.add_command(queue)



from datetime import datetime as dt
from os.path import realpath, split

from flask import (
    Flask, Blueprint, Request, Response, flash, g, jsonify,
    make_response as response, redirect, render_template as render, request,
    safe_join as join, session, url_for as to)

from flask_alembic import Alembic

from flask_assets import Bundle, Environment as Assets

from flask_babel import Babel

from flask_security import Security, cli as SecurityCLI

from .util import X, _, collapse, dump


__root_path = join('/', *split(realpath(__file__))[0].split('/'))


from . import (
    api as _api,
    auth as _auth,
    cache as _cache,
    cli as _cli,
    components as _components,
    config as _config,
    context as _context,
    db as _db,
    errors as _errors,
    filters as _filters,
    queues as _queues,
    sockets as _sockets,
    static as _static,
    templates as _templates,
    web as _web
)


api = _api.api
cache = _cache.cache
components = _components.components
config = _config.config
context = _context.context
db = _db.db
emit = _sockets.emit
env = _config.env
error = _errors.error
filter = _db.filter
forget = _cache.forget
memo = _cache.memo
model = _db.model
on = _db.on
queue = _queues.queue
register = _api.register
resource = _api.resource
schemas = _api.schemas
send = _sockets.send
socket = _sockets.socket
store = _cache.store
synonym = _db.synonym
web = _web.web
validates = _db.validates

Resource = _api.Resource

SecurityCLI.roles = None


class Application(Flask):

    def __init__(self, name='app', **params):

        super().__init__(name, **params)

        self.is_started = False

        _config.start(self)

        self.jinja_options = _config.jinja_options
        self.templates = self.jinja_env

        db.init_app(self)
        store.init_app(self)
        schemas.init_app(self)

        self.assets = Assets(self)
        self.locale = Babel(self)
        self.migrations = Alembic(self)

        self.users = _auth.Users(db, _auth.User, None)
        self.security = Security(self, self.users, register_blueprint=False)

        self.assets.auto_build = False
        self.assets.cache = self.config.cache_dir
        self.assets.directory = self.root_path
        self.assets.load_path = [self.root_path]
        self.assets.manifest = False
        self.assets.url = '/'

        self.templates.hamlish_enable_div_shortcut = True
        self.templates.hamlish_mode = 'indented'
        self.templates.filters.update(_context.filters)

    def start(self):

        if (self.is_started):
            raise ValueError('%s is already started' %
                             (self.import_name,))

        self.errorhandler(_errors.error)
        self.context_processor(_context.process)
        self.shell_context_processor(_context.process)

        _cli.start(self)
        _errors.start(self)
        _queues.start(self)
        _static.start(self)

        _api.start(self)
        _components.start(self)
        _web.start(self)

        self.is_started = True

        return self

# flake8: noqa



from flask_caching import Cache

store = Cache()

cache = store.cached

memo = store.memoize
forget = store.delete_memoized

clear = store.clear
delete = store.delete

get = store.get
set = store.set



from junior import Application, Bundle, join

application = Application()

application.assets.register('junior_js', Bundle(

    join('scripts', 'junior.js'),
    filters=('require', 'rjsmin'),
    output=join('src', 'junior', 'scripts', 'junior.js')
))



from junior import Application


class TestApplication:

    application = Application('junior')

    # Will the application start?
    def test_start(self):

        self.application.start()

    # Is the application's import name correct?
    def test_import_name(self):

        assert self.application.import_name == 'junior'



from . import dt, to
from .config import config
from .util import X, _


def asset(name):

    try:
        return to('static', filename=name)

    except AttributeError:
        return '%s/%s' % ('/static', name)


def title(text=None):

    return ('%s | %s' % (text, config.name) if text else config.name)


def process(key=None):

    return context[key] if key in context else None if key else context


filters = (_, X, asset, title)

filters = _(zip([f.__name__ for f in filters], filters))

context = _(
    dt=dt,
    now=dt.now(),
    config=config
)

context.update(filters)



from os import environ
from pathlib import Path

from flask import Config as FlaskConfig

from munch import Munch

from werkzeug.middleware.proxy_fix import ProxyFix

from yaml import safe_load as load

from . import __root_path, dt, join
from .util import _, collapse


class Config(Munch, FlaskConfig):
    pass


env = _()
config = _()
vendor = _()

babel = _()
postcss = _()

defaults = _(
    cache_default_timeout=60,
    cache_dir='.cache',
    cache_type='filesystem',
    database_url=None,
    flask_debug=False,
    proxies=0,
    secret_key=None,
    security_flash_messages=False,
    security_token_authentication_header='Authorization',
    sqlalchemy_track_modifications=False,
    static_folder='static',
    task_serializer='json'
)

celery_options = _(
    broker_url='filesystem://',
    broker_transport_options=_(
        data_folder_in=join(defaults.cache_dir, 'queue'),
        data_folder_out=join(defaults.cache_dir, 'queue'),
        data_folder_processed=defaults.cache_dir
    ),
    result_backend='file://%s' % (defaults.cache_dir,)
)

jinja_options = _(
    extensions=['hamlish_jinja.HamlishExtension',
                'junior.extensions.HamlBabel'],
    variable_start_string='{|',
    variable_end_string='|}'
)


with open(join(__root_path, 'config', 'babel.yaml')) as file:

    babel = _(load(file.read()))

try:
    with open(join('config', 'babel.yaml')) as file:

        babel.update(load(file.read()))

except (OSError, TypeError):
    pass


with open(join(__root_path, 'config', 'postcss.yaml')) as file:

    postcss = _(load(file.read()))

try:
    with open(join('config', 'postcss.yaml')) as file:

        postcss.update(load(file.read()))

except (OSError, TypeError):
    pass


try:
    with open(join('config', 'app.yaml')) as file:

        config.update(collapse(load(file.read())))

except (OSError, TypeError):
    pass


try:
    with open(join('config', 'env.yaml')) as file:

        env.update(collapse(load(file.read())))

except (OSError, TypeError):
    pass


try:
    with open(join('config', 'vendor.yaml')) as file:

        vendor.update(load(file.read()))

except (OSError, TypeError):
    pass


if 'api' not in config:
    config.api = _()

if 'version' not in config.api:
    config.api.version = 1

if 'updated_at' not in config.api:
    config.api.updated_at = dt.now()

for key in defaults:
    if key not in env:
        env[key] = environ.get(key.upper(), defaults[key])

env.babel_extra_args = ['--config-file',
                        join('.', env.cache_dir, 'babel.config.js')]

if 'security_password_salt' not in env:
    env.security_password_salt = environ.get('security_password_salt',
                                             env.secret_key)

if 'sqlalchemy_database_uri' not in env:
    env.sqlalchemy_database_uri = environ.get('sqlalchemy_database_uri',
                                              env.database_url)

env_upper = {key.upper(): env[key] for key in env}
env.update(env_upper)

for key in env_upper:
    environ[key] = str(env_upper[key])


def start(app):

    if 'name' not in config:
        config.name = app.import_name

    env.postcss_extra_args = ['--config', join(app.root_path, env.cache_dir)]
    env.POSTCSS_EXTRA_ARGS = env.postcss_extra_args

    app.config.update(env)

    if app.config['PROXIES']:
        app.wsgi_app = ProxyFix(app.wsgi_app,
                                x_for=app.config['PROXIES'],
                                x_proto=app.config['PROXIES'])

    app.config = Config(app.config)

    Path(env.cache_dir).mkdir(exist_ok=True)
    Path(join(env.cache_dir, '.empty')).touch()

    file = open(join(env.cache_dir, 'postcss.config.js'), 'w')
    file.write('module.exports = %s' % (postcss.toJSON(),))

    file = open(join(env.cache_dir, 'babel.cfg'), 'w')

    for format in babel:
        for source in babel[format]:

            file.write('[%s: %s]\n' % (format, source))

            for option in babel[format][source]:

                if isinstance(babel[format][source][option], list):
                    babel[format][source][option] = (
                        ','.join(babel[format][source][option]))

                file.write('%s = %s\n' % (option,
                                          babel[format][source][option]))

            file.write('\n')

    return app



from flask_security import (SQLAlchemyUserDatastore as UserDatastore,
                            UserMixin as WithUser, current_user, hash_password)

from . import request, response
from .api import Resource, register, resource
from .controls import allow_self_for_current_user, deny_all
from .db import WithTimestamps, db, filter, model, synonym
from .errors import BadRequest, Unauthorized, error
from .util import _


class Users(UserDatastore):

    def put(self, model):
        model.save()
        return model

    def delete(self, model):
        model.delete()


@register
@resource
@model
@filter('set', 'password', hash_password)
class User(db.Model, WithUser, WithTimestamps):

    class Meta:

        control = _(
            query=(deny_all, allow_self_for_current_user),
            update=allow_self_for_current_user,
        )

        exclude = ('token', 'password')

    id = db.Column(db.Integer, primary_key=True)

    name = db.Column(db.Text, unique=True, nullable=False)
    password = db.Column(db.Text, nullable=False)

    active = db.Column(db.Boolean, default=True, nullable=False)

    token = db.Column(db.Text, unique=True)
    fs_uniquifier = synonym('token')


@register
class TokenResource(Resource):

    class Meta:

        endpoint = 'token'

    def get(self, **params):

        return []

    def post(self, **params):

        data = _(request.get_json(force=True))

        try:
            user = User.query.filter(User.name == data.name).first()

        except AttributeError:
            e = error(BadRequest, 'Missing field: name')
            return {'error': e}, e.code

        if user is None:
            raise Unauthorized

        try:
            if user.verify_and_update_password(data.password):
                return {'id': user.get_auth_token()}

            raise Unauthorized

        except AttributeError:
            e = error(BadRequest, 'Missing field: password')
            return {'error': e}, e.code

    def delete(self, **params):

        current_user.set_uniquifier()
        return response('')



#!/usr/bin/env python


from setuptools import setup


with open('README.md', 'r') as file:
    readme = file.read()


setup(
    name='junior',
    version='0.2.2',
    url='https://gitlab.com/chriswhalen/junior',
    license='MIT',
    author='Chris Whalen',
    author_email='chris@chriswhalen.ca',
    description='An opinionated Flask project structure.',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=['junior'],
    package_data={'junior': ['config/*',
                             'scripts/*',
                             'templates/*']},
    package_dir={'': 'src'},
    platforms='any',
    entry_points={
        'flask.commands': [
            'run=junior.cli:run',
            'shell=junior.cli:shell'
        ]
    },
    install_requires=[
        'bcrypt~=3.1',
        'celery~=4.4',
        'flask~=1.1',
        'flask-alembic~=2.0',
        'flask-assets~=0.12',
        'flask-babel~=1.0',
        'flask-caching~=1.9',
        'flask-marshmallow~=0.12',
        'flask-restful~=0.3',
        'flask-security-too~=3.3',
        'flask-socketio~=4.3',
        'flask-sqlalchemy~=2.4',
        'gunicorn~=20.0',
        'hamlish-jinja~=0.3',
        'konch~=4.2',
        'marshmallow~=3.6',
        'marshmallow-sqlalchemy~=0.23',
        'munch~=2.3',
        'mysqlclient~=1.4',
        'psycopg2~=2.8',
        'pyyaml~=5.3',
        'readline~=6.2',
        'sqlalchemy~=1.3',
        'sqlalchemy-migrate~=0.13',
        'toml~=0.10',
        'webassets~=2.0'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



from re import sub
from sys import exc_info
from traceback import format_tb

from jinja2.exceptions import (
    FilterArgumentError, SecurityError, TemplateAssertionError, TemplateError,
    TemplateNotFound, TemplateRuntimeError, TemplateSyntaxError,
    TemplatesNotFound, UndefinedError)

from sqlalchemy.exc import (
    AmbiguousForeignKeysError, ArgumentError, CircularDependencyError,
    CompileError, DBAPIError, DataError, DatabaseError, DisconnectionError,
    IdentifierError, IntegrityError, InterfaceError, InternalError,
    InvalidRequestError, NoForeignKeysError, NoInspectionAvailable,
    NoReferenceError, NoReferencedColumnError, NoReferencedTableError,
    NoSuchColumnError, NoSuchModuleError, NoSuchTableError,
    ObjectNotExecutableError, OperationalError, ProgrammingError,
    ResourceClosedError, StatementError, TimeoutError, UnboundExecutionError,
    UnreflectableTableError, UnsupportedCompilationError)

from sqlalchemy.orm.exc import (
    ConcurrentModificationError, DetachedInstanceError, FlushError,
    LoaderStrategyException, MultipleResultsFound, NoResultFound,
    ObjectDeletedError, ObjectDereferencedError, StaleDataError,
    UnmappedClassError, UnmappedColumnError, UnmappedError,
    UnmappedInstanceError)

from webassets.exceptions import (
    BuildError, BundleError, EnvironmentError, FilterError)

from werkzeug.exceptions import (
    BadGateway, BadRequest, Conflict, ExpectationFailed, FailedDependency,
    Forbidden, GatewayTimeout, Gone, HTTPException, HTTPVersionNotSupported,
    ImATeapot, InternalServerError, LengthRequired, Locked, MethodNotAllowed,
    NotAcceptable, NotFound, NotImplemented, PreconditionFailed,
    PreconditionRequired, RequestEntityTooLarge, RequestHeaderFieldsTooLarge,
    RequestTimeout, RequestedRangeNotSatisfiable, ServiceUnavailable,
    TooManyRequests, Unauthorized, UnavailableForLegalReasons,
    UnprocessableEntity, UnsupportedMediaType)

from . import request
from .util import _


handlers = _()
debugging = False


class handle:

    def __init__(self, path):
        self.path = path

    def __call__(self, handler):
        handlers[self.path] = handler
        return handler


def debug():

    debug = _()
    ex = exc_info()

    if any(ex):

        debug.update(_(
            name=sub(r'([A-Z])', r' \1', ex[1].__class__.__name__).strip(),
            message=' '.join(str(ex[1]).split()).replace('"', "'"),
            traceback=[tb.strip().replace('"', "'").split(
                '\n') for tb in format_tb(ex[2])]
        ))

        for i, tb in enumerate(debug.traceback):
            debug.traceback[i] = [line.strip() for line in tb]

    return debug


def error(exception, message=None):

    error = _()

    if isinstance(exception, type):
        try:
            exception = exception()
        except Exception:
            pass

    if not isinstance(exception, HTTPException):
        exception = InternalServerError()

    error.code = exception.code
    error.name = sub(r'([A-Z])', r' \1', exception.__class__.__name__).strip()

    _debug = debug()
    if debugging and any(_debug):
        error.update(_debug)

    if message:
        error.message = message

    return error


def err(exception):

    for path in sorted(handlers.keys(), key=len)[::-1]:
        if request.path.startswith(path):
            return handlers[path](exception)


def start(app):

    from . import errors

    errors.debugging = app.debug

    return app



from pathlib import Path

from . import Bundle, __root_path, join
from .config import env
from .errors import BundleError


def start(app):

    app.assets._named_bundles = {}

    try:

        if Path(join('styles', 'app.sss')).is_file():

            app.assets.register('app_css', Bundle(

                join('styles', 'app.sss'),
                depends=(join('styles', '*.sss')),
                filters=('import_sss', 'handle_empty', 'postcss', 'import_css',
                         'cleancss'),
                output=join(app._static_folder, 'app.css')
            ))

        else:

            app.assets.register('app_css', Bundle(
                join(env.cache_dir, '.empty'),
                output=join(app._static_folder, 'app.css')
            ))

    except BundleError:
        pass

    try:

        if Path(join('scripts', 'app.js')).is_file():

            app.assets.register('app_js', Bundle(

                Bundle(join(__root_path, 'scripts', 'junior.js')),

                Bundle(join('scripts', 'app.js'),
                       depends=(join('scripts', '*.js'),),
                       filters=('require', 'rjsmin')),

                output=join(app._static_folder, 'app.js')
            ))

        else:

            app.assets.register('app_js', Bundle(

                join(__root_path, 'scripts', 'junior.js'),
                output=join(app._static_folder, 'app.js')
            ))

    except BundleError:
        pass

    return app



from os import listdir
from os.path import exists, isdir

from . import Blueprint, Bundle, join, render
from .errors import BundleError, TemplateNotFound, error, handle


components = Blueprint('components', __name__, template_folder='../components')

components.url_prefix = '/_/'


def defaults():

    from . import _components

    @_components.components.route('<path:path>')
    def __default(path):

        try:
            return render('%s.html' % (path,))

        except TemplateNotFound:
            return '<!-- not found: %s -->' % (path,), 404

    @handle(_components.components.url_prefix)
    @_components.components.errorhandler(Exception)
    def __error(exception):

        message = error(exception)
        return '<!-- error: %s -->' % (message.name.lower(),), message.code


def start(app):

    from . import _components

    names = []

    try:
        names = [c if isdir(join('components', c)) else None
                 for c in listdir('components')]

    except FileNotFoundError:
        pass

    for name in names:

        if not name:
            continue

        bundles = []
        path = join('components', name, '')

        if exists('%s%s' % (path, 'template.haml')):

            bundles.append(Bundle('%s%s' % (path, 'template.haml'),
                                  filters=('hamlish')))

        if exists('%s%s' % (path, 'style.sss')):

            bundles.append(Bundle('%s%s' % (path, 'style.sss'),
                                  filters=('handle_empty',
                                           'postcss',
                                           'wrap_style')))

        if exists('%s%s' % (path, 'script.js')):

            bundles.append(Bundle('%s%s' % (path, 'script.js'),
                                  filters=('require',
                                           'babel',
                                           'wrap_script')))

        try:
            app.assets.register(name,
                                *bundles,
                                filters='strip',
                                output=join('components',
                                            '%s.html' % (name,)))

        except BundleError:
            pass

    _components.defaults()

    app.register_blueprint(_components.components,
                           url_prefix=_components.components.url_prefix)

    return app



from pprint import pformat

from flask_babel import lazy_gettext, lazy_ngettext

from munch import DefaultMunch as Munch

from toml import dumps, loads


Munch.__name__ = '_'

Munch.toTOML = lambda self: dumps(self)
Munch.fromTOML = lambda data: Munch.fromDict(loads(data))


def _(_dict={}, **params):

    params.update(_dict)
    return Munch.fromDict(params, None)


def X(string, plural_string=None, count=1):

    if plural_string:
        return lazy_ngettext(string, plural_string, count)

    return lazy_gettext(string)


def collapse(data, delimiter='_'):

    values = _()

    for key in data:
        if isinstance(data[key], dict):
            inner_data = collapse(data[key], delimiter)
            for inner_key in inner_data:
                full_key = "%s%s%s" % (key, delimiter, inner_key)
                values[full_key] = inner_data[inner_key]
        else:
            values[key] = data[key]

    return values


def dump(*args, _print=True):

    string = '\n'

    for arg in args:

        try:
            if arg.__dict__:
                string += pformat(arg.__dict__)

            else:
                string += pformat(arg)

        except AttributeError:
            string += pformat(arg)

        string += '\n'

    if _print:
        print(string)

    return string
